@@ -1,15 +1,21 @@
 import os
+import re
 import subprocess
 import json
 import logging
-import re
 
 from subprocess import TimeoutExpired
-from typing import List
 from retry import retry
 from .download_bitwarden import DownloadBitwarden
 
 
+class BwErrorSamples:
+    MULTI_RESULTS = (
+        "More than one result was found. Try getting a specific object by "
+        "`id` instead. The following objects were found:"
+    )
+
+
 class BitwardenServerException(Exception):
     pass
 
@@ -26,9 +32,23 @@ class Bitwarden(object):
         """
         self.data = {}
         self.path_to_exe_file = "bw"
-        self.bitwarden_credentials = bitwarden_credentials
         self.session_key = ""
 
+        if bitwarden_credentials:
+            bitwarden_keys = ["client_id", "client_secret", "password"]
+            if all(k in bitwarden_credentials.keys() for k in bitwarden_keys):
+                os.environ["BW_CLIENTID"] = bitwarden_credentials["client_id"]
+                os.environ["BW_CLIENTSECRET"] = bitwarden_credentials["client_secret"]
+                self.bw_password = bitwarden_credentials["password"]
+            else:
+                raise ValueError("Missing client_id, client_secret, or password")
+        else:
+            if all((os.environ.get("BW_CLIENTID"), os.environ.get("BW_CLIENTSECRET"), os.environ.get("BW_PASSWORD"))):
+                logging.info("no bitwarden credentials, using environment variables")
+                self.bw_password = os.environ["BW_PASSWORD"]
+            else:
+                ValueError("No bitwarden credentials provided or in environment variables")
+
     def bitwarden_exe(self, *command):
         """
         Provide coma-separated command line arguments that you want to provide to bw CLI binary
@@ -37,26 +57,10 @@ class Bitwarden(object):
           - bw.bitwarden_exe('logout')
           - bw.bitwarden_exe(
             "unlock",
-            self.bitwarden_credentials["password"],
+            self.bw_password,
             "--raw",
             )
         """
-        env = {
-            **os.environ,
-        }
-
-        if self.bitwarden_credentials["client_id"] and self.bitwarden_credentials["client_secret"]:
-            env["BW_CLIENTID"] = str(self.bitwarden_credentials["client_id"])
-            env["BW_CLIENTSECRET"] = str(self.bitwarden_credentials["client_secret"])
-            logging.info("Using provided client_id and client_secret")
-        else:
-            logging.info("Using client_id and client_secret from env vars")
-
-        try:
-            if not env["BW_CLIENTID"] or not env["BW_CLIENTSECRET"]:
-                raise ValueError("Empty client_id or client_secret!")
-        except KeyError as e:
-            raise ValueError("Empty client_id or client_secret!") from e
 
         try:
             return subprocess.run(
@@ -67,7 +71,7 @@ class Bitwarden(object):
                 capture_output=True,
                 text=True,
                 timeout=180,
-                env=env,
+                env=os.environ,
             )
         except FileNotFoundError:
             self.path_to_exe_file = DownloadBitwarden.download_bitwarden()
@@ -79,48 +83,55 @@ class Bitwarden(object):
                 capture_output=True,
                 text=True,
                 timeout=180,
-                env=env,
+                env=os.environ,
             )
 
+    def bitwarden_logout(self):
+        return self.bitwarden_exe("logout")
+
     @retry((TimeoutExpired, BitwardenServerException), delay=5, tries=3)
     def bitwarden_login(self):
         """
         Performs login opeartion via BitWarden CLI
         Requires password / client_id / client_secret already set when creation Bitwarden instance
         """
-        self.bitwarden_exe("logout")
+        self.bitwarden_logout()
 
         bitwarden_app = self.bitwarden_exe(
             "login",
             "--apikey",
         )
 
-        if "You are logged in!" in bitwarden_app.stdout:
+        logging.info(bitwarden_app.stdout)
+        if bitwarden_app.returncode == 0:
             bitwarden_app = self.bitwarden_exe(
                 "unlock",
-                self.bitwarden_credentials["password"],
+                self.bw_password,
                 "--raw",
             )
 
-            if "Invalid master password" in bitwarden_app.stderr:
+            if bitwarden_app.returncode == 0:
+                self.session_key = bitwarden_app.stdout
+            else:
                 logging.error(f"STDOUT: {bitwarden_app.stdout}")
                 logging.error(f"STDERR: {bitwarden_app.stderr}")
                 raise ValueError("Invalid master password!")
-
-            if not re.search(r".{80,}", bitwarden_app.stdout):
-                logging.error(f"STDOUT: {bitwarden_app.stdout}")
-                logging.error(f"STDERR: {bitwarden_app.stderr}")
-                raise RuntimeError("No session key returned after login!")
-
-            self.session_key = bitwarden_app.stdout
         else:
             logging.error(f"STDOUT: {bitwarden_app.stdout}")
             logging.error(f"STDERR: {bitwarden_app.stderr}")
+            raise ValueError("Invalid bitwarden client_id or client_secret!")
 
-            if "invalid_client" in bitwarden_app.stderr:
-                raise ValueError("Invalid bitwarden client_id or client_secret!")
+    def sync(self):
+        bitwarden_app = self.bitwarden_exe(
+            "sync",
+            "--session",
+            self.session_key,
+        )
 
-            raise BitwardenServerException("Bitwarden Server error during login! Retrying...")
+        if bitwarden_app.returncode != 0 or bitwarden_app.stdout != "Syncing complete.":
+            logging.error(f"STDOUT: {bitwarden_app.stdout}")
+            logging.error(f"STDERR: {bitwarden_app.stderr}")
+            raise RuntimeError("Unable to sync bitwarden")
 
     def get_credentials(self, user_credentials_name):
         """
@@ -157,54 +168,72 @@ class Bitwarden(object):
         )
 
         print("Getting bitwarden data...")
-        bitwarden_app = self.bitwarden_exe(
-            "list",
-            "items",
-            "--session",
-            self.session_key,
-        )
-
-        if not bitwarden_app.stdout:
-            logging.error(f"STDERR: {bitwarden_app.stderr}")
-            raise RuntimeError("Empty result for 'bw list items' command! This user has no items associated")
-
-        bitwarden_items = json.loads(bitwarden_app.stdout)
-        items_not_found: List[str] = []
-        for credentials_key, credentials_name in data.items():
-            for bw_item in bitwarden_items:
-                if credentials_name == bw_item["name"]:
-                    self.data[credentials_key] = {
-                        "login": bw_item["login"]["username"],
-                        "password": bw_item["login"]["password"],
-                        "url": bw_item["login"]["uris"][0]["uri"] if "uris" in bw_item["login"] else "",
-                    }
-
-                    if bw_item["login"]["totp"] is None:
-                        self.data[credentials_key]["otp"] = ""
-                    else:
-                        bitwarden_app = self.bitwarden_exe(
-                            "get",
-                            "totp",
-                            bw_item["id"],
-                            "--session",
-                            self.session_key,
-                        )
-                        self.data[credentials_key]["otp"] = bitwarden_app.stdout
-                    if "fields" in bw_item:
-                        for field in bw_item["fields"]:
-                            self.data[credentials_key][field["name"]] = field["value"]
-
-                    break
-            else:
-                items_not_found.append(credentials_name)
-
-        if items_not_found:
+        if bitwarden_app.stdout != "Syncing complete.":
             logging.error(f"STDOUT: {bitwarden_app.stdout}")
             logging.error(f"STDERR: {bitwarden_app.stderr}")
-            raise BitwardenServerException(
-                "Invalid bitwarden collection or key name or no access to collection for this user! Items not found "
-                f"include {', '.join(items_not_found)}"
+            logging.warning("Unable to sync bitwarden")
+
+        for credentials_key, credentials_name in data.items():
+            command = ["get", "item", credentials_name, "--session", self.session_key]
+            logging.info(f"running {command=}")
+            bitwarden_app = self.bitwarden_exe(*command)
+            if bitwarden_app.returncode != 0 or bitwarden_app.stdout == "":
+                logging.error(f"{bitwarden_app.returncode=}")
+                logging.error(f"STDOUT: {bitwarden_app.stdout}")
+                logging.error(f"STDERR: {bitwarden_app.stderr}")
+
+                if BwErrorSamples.MULTI_RESULTS in bitwarden_app.stderr:
+                    logging.info("attempting to parse and request IDs")
+                    item_id_re = r"((?:[\d\w]+-){4}[\d\w]+)"
+                    item_ids = re.findall(item_id_re, bitwarden_app.stderr)
+                    if item_ids:
+                        for item_id in item_ids:
+                            command = ["get", "item", item_id, "--session", self.session_key]
+                            logging.info(f"running {command=}")
+                            bitwarden_app = self.bitwarden_exe(*command)
+                            item_dict = self._parse_credential_response(bitwarden_app.stdout, credentials_key)
+                            if item_dict["name"] == credentials_name or item_dict["id"] == credentials_name:
+                                self.data[credentials_key] = item_dict
+                                break
+                            # if item
+                        else:
+                            raise BitwardenServerException(f"Failed to udentify exact match for  {credentials_name=}")
+                else:
+                    raise BitwardenServerException(
+                        f"Invalid bitwarden collection or key name or no access to collection for this user! "
+                        f"{credentials_name=} not found"
+                    )
+
+            item_dict = self._parse_credential_response(bitwarden_app.stdout, credentials_key)
+            self.data[credentials_key] = item_dict
+
+    def _parse_credential_response(self, bitwarden_stdout: str, credentials_key):
+        bw_item = json.loads(bitwarden_stdout)
+        item = {
+            "id": bw_item["id"],
+            "name": bw_item["name"],
+            "login": bw_item["login"]["username"],
+            "password": bw_item["login"]["password"],
+            "url": bw_item["login"]["uris"][0]["uri"] if "uris" in bw_item["login"] else "",
+        }
+
+        if bw_item["login"]["totp"] is None:
+            item["otp"] = ""
+        else:
+            bitwarden_app = self.bitwarden_exe(
+                "get",
+                "totp",
+                bw_item["id"],
+                "--session",
+                self.session_key,
             )
+            item["otp"] = bitwarden_app.stdout
+
+        if "fields" in bw_item:
+            for field in bw_item["fields"]:
+                item[field["name"]] = field["value"]
+
+        return item
 
     @retry((TimeoutExpired, BitwardenServerException), delay=5, tries=3)
     def get_attachment(self, item_name, file_name, output_folder=os.getcwd()):
@@ -231,7 +260,7 @@ class Bitwarden(object):
             os.mkdir(output_folder)
 
         print("Syncing bitwarden data...")
-        bitwarden_app = self.bitwarden_exe(
+        self.bitwarden_exe(
             "sync",
             "--session",
             self.session_key,
@@ -246,7 +275,7 @@ class Bitwarden(object):
             self.session_key,
         )
 
-        if bitwarden_app.stderr:
+        if bitwarden_app.returncode != 0:
             logging.error(f"STDOUT: {bitwarden_app.stdout}")
             logging.error(f"STDERR: {bitwarden_app.stderr}")
             if "A 404 error occurred while downloading the attachment" in bitwarden_app.stderr:
